#include <stdlib.h>
#include <stdio.h>
#include "gfx.h"
#include "network.h"
#include "draw-net.h"

static network_t *NW;

typedef struct { int x,y,z; } triplet_t;
#define DENSMAX 500
static triplet_t dens[DENSMAX];
static int ndens;


static void show_stats()
{
  char buf[200];
  sprintf(buf,"%d gates %d levels",
      NW->ngates, NW->gates[NW->ngates-1]->level+1);
  gfx_set_text(Wstats,buf);
}

static void show_network(int siz)
{
  int w,h;
  unsigned int *data;
  if (!NW) return ;
  draw_network(NW, 30,30,7,3,&w,&h,&data);
  if (siz)
    gfx_canvas_setsize(Wcanvas, w,h);
  gfx_canvas_put(Wcanvas,0,0,w,h,data,w,h,0,0);
  free(data);
  show_stats();
}

static void replace_network(network_t *WW)
{
  network_free(NW);
  NW= WW;
  show_network(1);
}

static char *txtfiles[]=
{
  "Text Files (.txt)",
  NULL
};

#if 0
static int choose_prefix_existing_pair
  (network_t *NW,int *Ra,int *Rb,int nr,unsigned char *remaining,
                                        unsigned char *visited)
{
  // First of all, we want to eliminate single-gate levels. Let's see
  // if we can find any.
  int i;
  compex_t *CE;
  for(i=0;i<NW->ngates;i++)
  {
    CE= NW->gates[i];
    if (CE->level==0) continue;
    if (NW->gates[i-1]->level!=CE->level &&
        (i==NW->ngates-1 || NW->gates[i+1]->level!=CE->level)
        && !visited[CE->A] && !visited[CE->B])
    {
          *Ra= CE->A;
          *Rb= CE->B;
          return 1;
    }
  }
  return 0;
}
#endif

static void choose_prefix(network_t *NW,int *Ra, int *Rb)
{
  unsigned char remaining[200];
  unsigned char visited[200];
  int i,j,nr;
  for(i=0;i<NW->ninputs;i++) visited[i]= 0; 
  for(i=0;i<NW->ngates;i++)
    if (NW->gates[i]->level!=0) break;
    else { visited[NW->gates[i]->A]= visited[NW->gates[i]->B]= 1; }
  nr= 0;
  for(i=0;i<NW->ninputs;i++)
    if (!visited[i]) remaining[nr++]= i;

  if (nr<2)
  {
    i= rand() % NW->ninputs;
    for(j=i;j==i;) j= rand()%NW->ninputs;
    *Ra= i; *Rb= j;
    return ;
  }
#if 0
  if (choose_prefix_existing_pair(NW,Ra,Rb,nr,remaining,visited)) return ;
#endif

  i= rand() % nr;
  *Ra= remaining[i];
  remaining[i]= remaining[nr-1];
  nr--;
  if (nr==1) { *Rb= remaining[0]; return ; }
  j= rand() % nr;
  *Rb= remaining[j];
}

static void get_level_avail(gatelist_t *GL,int level,int *Rna,int *avail)
{
  int i;
  char visited[64];
  int N;
  int na;
  gate_t *g;
  N= GL->ninputs;
  for(i=0;i<N;i++) visited[i]= 0;
  for(i=0;i<GL->ngates;i++)
  {
    g= GL->gates+i;
    if (g->level==level) visited[g->A]= visited[g->B]= 1;
    else if (g->level>level) break;
  }
  na= 0; 
  for(i=0;i<N;i++)
    if (!visited[i]) avail[na++]= i;
  *Rna= na;
}

static network_t *make_network_with_insertion
  (gatelist_t *GL,int level,int a,int b)
{
  int i;
  network_t *WW;
  gate_t *g;
  WW= network_new(GL->ninputs);
  for(i=0;i<GL->ngates;i++)
  {
    g= GL->gates+i;
    if (g->level==level)
    {
      network_append_gate(WW, a,b);
      level= -1;
    }
    network_append_gate(WW, g->A, g->B);
  }
  return WW;
}

static void remove_something()
{
  gatelist_t *GL;
  int i,j;
  network_t *WW;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  for(i=GL->ngates-1;i>=0;i--)
  {
    WW= network_new(GL->ninputs);
    for(j=0;j<GL->ngates;j++)
     if (j!=i) network_append_gate(WW,GL->gates[j].A,GL->gates[j].B);
    if (network_sorts(WW))
    {
      for(j=0;j<GL->ngates;j++)
        if (j==i) NW->gates[j]->color=0xff0000;
             else NW->gates[j]->color=0x000000;
      show_network(0);
      network_free(WW);
      goto done;
    }
    network_free(WW);
  }
done:
  gatelist_free(GL);
}

static int densify()
{
  gatelist_t *GL;
  int i,j,k,A,B,L;
  network_t *WW;
  int avail[50]; int navail;

  if (!NW) return 0;
  network_to_gate_list(NW,&GL);

  ndens= 0;
  for(i=1;i<=GL->gates[GL->ngates-1].level;i++)
  {
    get_level_avail(GL, i, &navail, avail);
    for(j=0;j<navail;j++)
     for(k=j+1;k<navail;k++)
      { if (ndens==DENSMAX) goto select_again;
        dens[ndens].x= avail[j];
        dens[ndens].y= avail[k];
        dens[ndens].z= i; 
        ndens++; }
  }

select_again:
  if (ndens==0) goto fail;
  k= rand()%ndens;
  A= dens[k].x;
  B= dens[k].y;
  L= dens[k].z;
  dens[k]= dens[ndens-1];
  ndens--;

  WW= make_network_with_insertion(GL, L, A,B);
  if (network_sorts(WW))
  {
    gatelist_free(GL);
    replace_network(WW);
    return 1;
  }
  network_free(WW);
  goto select_again;

fail:
  return 0;
}

static int normalize_gatelist_reverse(gatelist_t *GL)
{
  int change, nextlevel, curlevel,i,j,N;
  char L[200];
  gate_t t, *g;

  change= 0;
  N= GL->ninputs;

  for(nextlevel=GL->gates[GL->ngates-1].level;nextlevel>0;nextlevel--)
  {
    for(i=0;i<N;i++) L[i]= 0;
    for(i=GL->ngates-1;i>=0;i--)
    {
       g= GL->gates+i;
       if (g->level>nextlevel) continue;
       if (g->level==nextlevel)
          L[g->A]= L[g->B]= 1;
       else break;
    }

    curlevel= nextlevel-1;
    for(j=i;j>=0;j--)
    {
      g= GL->gates+j;
      if (g->level!=curlevel) break;
      if (!L[g->A] && !L[g->B])
      {
        g->level= nextlevel;
        L[g->A]= L[g->B]= 1;
        t= GL->gates[i];
        GL->gates[i]= GL->gates[j];
        GL->gates[j]= t;
        i--;
        change= 1;
      }
    }
  }
  return change;
}


static int normalize_gatelist(gatelist_t *GL)
{
  int change, prevlevel, curlevel,i,j,N;
  char L[200];
  gate_t t, *g;

  change= 0;
  N= GL->ninputs;

  for(prevlevel=0;prevlevel<GL->gates[GL->ngates-1].level-1;prevlevel++)
  {
    for(i=0;i<N;i++) L[i]= 0;
    for(i=0;i<GL->ngates;i++)
    {
       g= GL->gates+i;
       if (g->level<prevlevel) continue;
       if (g->level==prevlevel)
          L[g->A]= L[g->B]= 1;
       else break;
    }

    curlevel= prevlevel+1;
    for(j=i;j<GL->ngates;j++)
    {
       g= GL->gates+j;
       if (g->level!=curlevel) break;
       if (!L[g->A] && !L[g->B])
       {
         L[g->A]= L[g->B]= 1;
         t= GL->gates[i];
         GL->gates[i]= GL->gates[j];
         GL->gates[j]= t;
         GL->gates[i].level= prevlevel;
         i++;
         change= 1;
       }
    }
  }
  return change;
}

static void print_gatelist_levels(gatelist_t *GL)
{
  int i;
  for(i=0;i<GL->ngates;i++) printf("%2d ", GL->gates[i].level);
  printf("\n");
}

static void do_normalize(int right)
{
  gatelist_t *GL;
  if (!NW) return;

  network_to_gate_list(NW,&GL);

  if (right) while(normalize_gatelist_reverse(GL));
       else  while(normalize_gatelist(GL));

  replace_network(network_from_gate_list(GL,right));
  gatelist_free(GL);
}

static void choose_prefix_for_last()
{
  int i,j,k;
  int N;
  int lastA, lastB, ll;
  int nu, u[500];
  network_t *WW;
  compex_t *CE;
  int maxnu;
  network_t *best;

  maxnu= 1;
  best= NULL;

  N= NW->ninputs;
  lastA= NW->gates[NW->ngates-1]->A;
  lastB= NW->gates[NW->ngates-1]->B;

  for(i=0;i<N;i++)
   for(j=i+1;j<N;j++)
   {
     WW= network_new(N);
     network_append_gate(WW, i,j);
     for(k=0;k<NW->ngates;k++)
       network_append_gate(WW, NW->gates[k]->A, NW->gates[k]->B);
     network_all_useless_gates(WW,&nu,u);
     if (nu==0)
     {
       network_free(WW);
       continue;
     }
     network_levels(WW);
     ll= WW->gates[WW->ngates-1]->level;
     for(k=0;k<nu;k++)
     {
       CE= WW->gates[u[k]];
       if (CE->level==ll && CE->A==lastA && CE->B==lastB)
       {
          replace_network(WW);
          if (best) network_free(best);
          return ;
       }
     }
     if (nu>maxnu) { maxnu= nu; if (best) network_free(best);
                                best= WW; }                     
     else network_free(WW);
   }

   if (best) replace_network(best);
}

static void flip_and_show_useless()
{
  network_t *WW;
  gatelist_t *GL;
  int i;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  WW= network_new(GL->ninputs);
  for(i=GL->ngates-1;i>=0;i--)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  gatelist_free(GL);
  if (!network_sorts(WW))
  {
    network_free(WW);
    return ;
  }
  i= network_find_useless_gate(NW);
  if (i!=-1) WW->gates[i]->color= 0xff0000;
  replace_network(WW);
}

static void reproduce()
{
  network_t *WW;
  gatelist_t *GL;
  int i;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  WW= network_new(GL->ninputs);
  for(i=0;i<GL->ngates;i++)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  for(i=GL->ngates;i>=0;i--)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  gatelist_free(GL);
  replace_network(WW);
}

static void swap_simple()
{
  gatelist_t *GL;
  gate_t t;
  int i,prevlev;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);

  prevlev= -1;
  for(i=GL->ngates-1;i>0;i--)
  {
    if (GL->gates[i].level!=prevlev && GL->gates[i-1].level!=GL->gates[i].level)
    {
      t= GL->gates[i];
      GL->gates[i]= GL->gates[i-1];
      GL->gates[i-1]= t;    
      replace_network(network_from_gate_list(GL,0));
      gatelist_free(GL);
      return ;
    }
    prevlev= GL->gates[i].level;
  }
  gatelist_free(GL);
}

int main(int argc,char **argv)
{
  char *fn;
  int i,j,u[400],nu;
  network_t *WW;

  if (argc>2) NW= network_from_cmdline(argv[1],argv[2]);
         else NW= network_from_file(argv[1]);
  gfx_init(NULL, 1000, 800, "Sorting Network Tool");
  gfx_SCREEN_0();
  show_network(1);
  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT:
       goto done;
    case Wuseless:
       if (!NW) continue;
       network_all_useless_gates(NW,&nu,u);
       if (nu==0) continue;
       for(i=0;i<nu;i++)
          NW->gates[u[i]]->color= 0xff0000;
       show_network(0);
       break;
    case Wremove:
       if (!NW) continue;
#if 0
       nu= 0;
       for(i=0;i<NW->ngates;i++)
         if (NW->gates[i]->color==0xff0000) u[nu++]= i;
       if (nu==0) continue;
       i= u[nu/2];
#endif
       WW= network_new(NW->ninputs);
       for(j=0;j<NW->ngates;j++)
         if (NW->gates[j]->color!=0xff0000)
           network_append_gate(WW,NW->gates[j]->A,NW->gates[j]->B);
       replace_network(WW);
       break;
    case Wrandomprefix:
       if (!NW) continue;
       WW= network_new(NW->ninputs);
       choose_prefix(NW, &i,&j);
       network_append_gate(WW,i,j);
       for(j=0;j<NW->ngates;j++)
         network_append_gate(WW,NW->gates[j]->A,NW->gates[j]->B);
       replace_network(WW);
       break;
     case Wnormalize: do_normalize(0); break;
     case Wrnormalize: do_normalize(1); break;
     case Wtryremove: remove_something(); break;
   
     case Wsave:
       if (!NW) continue;
       fn= gfx_select_file(1,NULL,txtfiles);
       if (fn) network_to_file(NW, fn);
       break;
     case Wdensify: while(densify()); break;
     case Wsorts: if (NW && network_sorts(NW)) gfx_set_text(Wstats,"Sorts");
				 else gfx_set_text(Wstats,"Doesn't sort");
          break;
     case Wremovefinal: 
          choose_prefix_for_last(); break;
     case Wreproduce: reproduce(); break;
     case Wflipuseless: flip_and_show_useless(); break;
     case Wswap: 
            swap_simple(); break;
    }
  }
done:
  return 0;
}
