#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "network.h"

static int factorial(int N)
{
  static int table[21];
  if (!table[0])
  {
    int i;
    table[0]= 1;
    for(i=1;i<20;i++) table[i]= i*table[i-1];
  }
  return table[N];
}

static int n_combinations(int N,int M)
{
  return factorial(N)/factorial(M)/factorial(N-M);
}

static int expterm_compare(const void *pA, const void *pB)
{
  input_t A,B;
  A= *(const input_t*)pA;
  B= *(const input_t*)pB;
  if (A>B) return 1;
  if (A<B) return -1;
  return 0;
}

static void exp_unify(exp_t *E)
{
  size_t L,i,j;
  L= E->len;
  qsort(E->data, L, sizeof(E->data[0]), expterm_compare);
  if (L<2) return ;
  j= 0;
  for(i=1;i<L;i++)
    if (E->data[i]!=E->data[j])
      E->data[++j]= E->data[i];
  E->data[++j]= 0;
  E->len= j;
}

static void exp_simplify(exp_t *E)
{
  size_t i,j;
  size_t removed;
  input_t *R;
  size_t lR;

  exp_unify(E);

  removed= 0;
    // RULE-1
  lR= E->len;
  R= E->data;
  for(i=0;i<lR-1;i++)
    for(j=i+1;j<lR;j++)
    {
      input_t K;
      K= R[i] & R[j];
      if (K==R[i]) 
      {
         R[j]= R[lR-1];
         lR--;
         j--;
         removed++;
         if (i>=lR-1) goto done;
      }
      else if (K==R[j])
      {
         R[i]= R[lR-1];
         lR--;
         i--;
         j= lR+1;
         removed++;
      }
    }
done:
  R[lR]= 0;
  E->len= lR;
  if (removed) exp_unify(E);
  if (removed>=lR/2 && removed>=1000)
     R= realloc(R, (lR+1)*sizeof(input_t));
  E->data= R;
}

static void exp_allocdata(exp_t *E)
{
  E->data= malloc(sizeof(E->data[0])*(E->len+1));
}

exp_t *exp_simple(int inputindex)
{
  exp_t *E;
  E= malloc(sizeof(*E));
  E->len= 1;
  exp_allocdata(E);
  E->data[0]= (1<<inputindex);
  E->data[1]= 0;
  E->count= 1;
  return E;
}

void exp_free(exp_t *E)
{
  free(E->data);
  free(E);
}

exp_t *exp_min(exp_t *A,exp_t *B)
{
  exp_t *C;
  int i,j,k;
  C= malloc(sizeof(*C));
  C->len= A->len*B->len;
  exp_allocdata(C);
  k= 0;
  for(i=0;i<A->len;i++)
   for(j=0;j<B->len;j++)
     C->data[k++]= A->data[i] | B->data[j];
  C->data[k]= 0;
  exp_simplify(C);
  return C;
}

exp_t *exp_max(exp_t *A,exp_t *B)
{
  exp_t *C;
  int i,j,k;
  C= malloc(sizeof(*C));
  C->len= A->len+B->len;
  exp_allocdata(C);
  k= 0;
  for(i=0;i<A->len;i++) C->data[k++]= A->data[i];
  for(j=0;j<B->len;j++) C->data[k++]= B->data[j];
  C->data[k]= 0;
  exp_simplify(C);
  return C;
}

network_t *network_new(int ninputs)
{
  network_t *NW;
  int i;
  NW= calloc(1,sizeof(*NW));
  NW->ninputs= ninputs;
  NW->sgates= 50;
  NW->gates= malloc(sizeof(NW->gates[0])*NW->sgates);
  NW->G= malloc(sizeof(NW->G[0])*(NW->sgates+1));
  NW->G[0]= malloc(sizeof(NW->G[0][0])*NW->ninputs);
  for(i=0;i<ninputs;i++)
    NW->G[0][i]= exp_simple(i);
  return NW;
}

void network_append_gate(network_t *NW, int A,int B)
{
  compex_t *CE;
  int i;

  if (A<0 || A>=NW->ninputs) return ;
  if (B<0 || B>=NW->ninputs) return ;
  if (A==B) return ;

  if (A>B) { int t; t= A; A=B; B= t; }

  CE= calloc(1,sizeof(*CE));
  CE->A= A;
  CE->B= B;
  CE->min= exp_min(NW->G[NW->ngates][A], NW->G[NW->ngates][B]);
  CE->max= exp_max(NW->G[NW->ngates][A], NW->G[NW->ngates][B]);

  if (NW->ngates==NW->sgates)
  {
    NW->sgates+= NW->sgates/4;
    NW->gates= realloc(NW->gates, sizeof(NW->gates[0])*NW->sgates);
    NW->G= realloc(NW->G, sizeof(NW->G[0])*(NW->sgates+1));
  }

  NW->gates[NW->ngates]= CE;
  NW->ngates++;
  NW->G[NW->ngates]= malloc(sizeof(NW->G[NW->ngates][0])*NW->ninputs);
  for(i=0;i<NW->ninputs;i++)
    if (i==A)      NW->G[NW->ngates][i]= CE->min;
    else if (i==B) NW->G[NW->ngates][i]= CE->max;
    else           NW->G[NW->ngates][i]= NW->G[NW->ngates-1][i];
}

static int set_bits(input_t k)
{
  int V;
  V= 0;
  while(k)
  {
   if (k&1) V++;
   k>>= 1;
  }
  return V;
}

int network_sorts(network_t *NW)
{
  //
  // Now I'm confident in my new method so I'm going to
  // remove the old method completely
  //
  exp_t **G;
  exp_t *E;
  int N,z,i;
  G= NW->G[NW->ngates];
  N= NW->ninputs;
  for(z=0;z<N;z++)
  {
    int p,compp,pp;
    E= G[z];
    p= N-1-z;
    pp= p+1;
    compp= n_combinations(N,pp);
    if (E->len!=compp) return 0; 
    for(i=0;i<compp;i++)
      if (set_bits(E->data[i])!=pp) return 0;
  }
  return 1;
}

void network_layout_for_drawing(network_t *NW)
{
  int i,x,j;
  int A,B,N;
  char visited[200];

  N= NW->ninputs;
  for(j=0;j<N;j++) visited[j]= 0;
  network_levels(NW);

  x= 0;
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    for(j=A;j<=B;j++) if (visited[j]) break;
    if (j<=B || 
          (i>0 && NW->gates[i]->level!=NW->gates[i-1]->level))
    {
      x++;
      for(j=0;j<N;j++) visited[j]= 0;
    }
    NW->gates[i]->x= x;
    for(j=A;j<=B;j++) visited[j]= 1;
  }
}



network_t *network_from_cmdline(char *a1,char *a2)
{
  int N,A,B;
  network_t *NW;
  char *end;

  N= atoi(a1);
  NW= network_new(N);

  while(1)
  {
    A= strtol(a2,&end,10);
    if (end==a2) break;
    a2= end+1;
    B= strtol(a2,&end,10);
    if (end==a2) break;
    a2= end+1;
    network_append_gate(NW, A,B);
  }
  return NW;
}

network_t *network_from_binary(int len, unsigned char *data)
{
  int N;
  network_t *NW;

  N= data[0];
  data++;
  len--;
  NW= network_new(N);

  while(len>=2)
  {
    network_append_gate(NW, data[0],data[1]);
    data+= 2;
    len-= 2;
  }
  return NW;
}

void network_to_binary(network_t *NW,int *Rlen, unsigned char **Rdata)
{
  int len;
  unsigned char *data;
  int i,o;

  len= 1 + NW->ngates*2;
  data= malloc(len);

  o= 0;
  data[o++]= NW->ninputs;
  for(i=0;i<NW->ngates;i++)
  {
    data[o++]= NW->gates[i]->A;
    data[o++]= NW->gates[i]->B;
  }
  *Rlen= len;
  *Rdata= data;
}

network_t *network_from_file(char *fn)
{
  FILE *f;
  int N,A,B;
  network_t *NW;

  f= fopen(fn,"rb");
  if (!f)  goto fail;

  if (fscanf(f,"%d",&N)!=1) goto fail;
  if (N<2 || N>20) goto fail;
  NW= network_new(N);
  while(fscanf(f,"%d%d",&A,&B)==2) 
    network_append_gate(NW,A,B);
  fclose(f);
  return NW;

fail:
  NW= network_new(2);
  network_append_gate(NW,0,1);
  return NW;
}

void network_to_file(network_t *NW,char *fn)
{
  FILE *f;
  int i;
  f= fopen(fn,"wb");
  if (!f) return ;
  fprintf(f,"%d\n",NW->ninputs);
  for(i=0;i<NW->ngates;i++)
   fprintf(f,"%d %d\n",NW->gates[i]->A,NW->gates[i]->B);
  fclose(f);
}

int exp_compare(exp_t *A,exp_t *B)
{
  size_t i;
  int L;
   // if B[i] becomes zero before A[i], A[i]>B[i] will be true
   // and we return 1.
   // if A[i] becomes zero before B[i], then the loop stops
   // and we check for B[i].
   // if they both become zero at the same i, then we return 0.
  L= A->len < B->len ? A->len : B->len;
  for(i=0;i<L;i++)
    if (A->data[i]<B->data[i]) return -1;
    else if (A->data[i]>B->data[i]) return 1;
  if (A->len==B->len) return 0;
  if (A->len>B->len) return 1;
  return -1;
}

char *exp_to_string(exp_t *A)
{
  char *d;
  int L,i,j;
  input_t Z;
  const char *digits= "0123456789abcdefghijklmnopqr";

  d= NULL;
again:
  L= 0;
  for(i=0;i<A->len;i++)
  {
    if (i) { if (d) d[L]= '+'; L++; }
    Z= A->data[i];
    for(j=0;Z;j++,Z>>=1)
     if (Z&1) { if (d) d[L]= digits[j]; L++; }
  }
  if (!d) { d= malloc(L+1); goto again; }
  d[L]= 0;
  return d;
}

static int useless_gate(network_t *NW,int idx)
{
  compex_t *CE;
  CE= NW->gates[idx];
  if (exp_compare(CE->max, NW->G[idx][CE->B])==0) return 1;
  // skipping the rename-useless stuff for now
  return 0;
}

int network_find_useless_gate(network_t *NW)
{
  int i;
  for(i=0;i<NW->ngates;i++)
    if (useless_gate(NW,i)) return i;
  return -1;
}

int network_find_useless_gate_reverse(network_t *NW)
{
  int i;
  for(i=NW->ngates-1;i>=0;i--)
    if (useless_gate(NW,i)) return i;
  return -1;
}

int network_levels(network_t *NW)
{
  char visited[200];
  int i,j,N;
  int level;
  compex_t *CE;

  N= NW->ninputs;
  for(j=0;j<N;j++) visited[j]= 0;
  level= 0;

  if (NW->direction)
  {
     for(i=NW->ngates-1;i>=0;i--)
     {
       CE= NW->gates[i];
       if (visited[CE->A] || visited[CE->B])
       {
         level++;
         for(j=0;j<N;j++) visited[j]= 0;
       }
       CE->level= level;
       visited[CE->A]= visited[CE->B]= 1;
     }

     for(i=0;i<NW->ngates;i++)
     {
       CE= NW->gates[i];
       CE->level= level - CE->level;
     }
  }
  else
  {
     for(i=0;i<NW->ngates;i++)
     {
       CE= NW->gates[i];
       if (visited[CE->A] || visited[CE->B])
       {
         level++;
         for(j=0;j<N;j++) visited[j]= 0;
       }
       CE->level= level;
       visited[CE->A]= visited[CE->B]= 1;
     }
  }
  return level+1;
}

void network_free(network_t *NW)
{
  int i;
  for(i=0;i<=NW->ngates;i++)
    free(NW->G[i]);
  free(NW->G);
  for(i=0;i<NW->ngates;i++)
  {
    exp_free(NW->gates[i]->min);
    exp_free(NW->gates[i]->max);
    free(NW->gates[i]);
  }
  free(NW->gates);
  free(NW);
}

void network_all_useless_gates(network_t *NW,int *Rn, int *Ru)
{
  int i;
  int n;
  n= 0;
  for(i=0;i<NW->ngates;i++)
   if (useless_gate(NW,i)) Ru[n++]= i;
  *Rn= n;
}


void network_to_gate_list(network_t *NW,gatelist_t **rGL)
{
  int i;
  gatelist_t *GL;
  GL=malloc(sizeof(*GL));
  GL->ninputs= NW->ninputs;
  GL->ngates= NW->ngates;
  GL->gates=malloc(sizeof(GL->gates[0])*GL->ngates);
  for(i=0;i<GL->ngates;i++)
  {
    GL->gates[i].A= NW->gates[i]->A;
    GL->gates[i].B= NW->gates[i]->B;
    GL->gates[i].level= NW->gates[i]->level;
    GL->gates[i].x= NW->gates[i]->x;
  }
  *rGL= GL;
}

network_t* network_from_gate_list(gatelist_t *GL,int right)
{
  network_t *NW;
  int i;
  NW= network_new(GL->ninputs);
  NW->direction= right;
  for(i=0;i<GL->ngates;i++)
    network_append_gate(NW, GL->gates[i].A, GL->gates[i].B);
  return NW;
}

void gatelist_free(gatelist_t *GL)
{
  free(GL->gates);
  free(GL);
}
